@@ -1,19 +1,10 @@
+# SPDX-License-Identifier: Apache-2.0
 # Copyright 2017 The Meson development team
 
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-
-#     http://www.apache.org/licenses/LICENSE-2.0
-
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
-
 """Code that creates simple startup projects."""
 
+from __future__ import annotations
+
 from pathlib import Path
 from enum import Enum
 import subprocess
@@ -22,22 +13,37 @@ import sys
 import os
 import re
 from glob import glob
-from mesonbuild import mesonlib
-from mesonbuild.environment import detect_ninja
-from mesonbuild.templates.samplefactory import sameple_generator
 import typing as T
 
+from mesonbuild import build, mesonlib, mlog
+from mesonbuild.coredata import FORBIDDEN_TARGET_NAMES
+from mesonbuild.environment import detect_ninja
+from mesonbuild.templates.mesontemplates import create_meson_build
+from mesonbuild.templates.samplefactory import sample_generator
+
 if T.TYPE_CHECKING:
     import argparse
 
-'''
-we currently have one meson template at this time.
-'''
-from mesonbuild.templates.mesontemplates import create_meson_build
+    from typing_extensions import Protocol, Literal
+
+    class Arguments(Protocol):
+
+        srcfiles: T.List[Path]
+        wd: str
+        name: str
+        executable: str
+        deps: str
+        language: Literal['c', 'cpp', 'cs', 'cuda', 'd', 'fortran', 'java', 'rust', 'objc', 'objcpp', 'vala']
+        build: bool
+        builddir: str
+        force: bool
+        type: Literal['executable', 'library']
+        version: str
+
 
 FORTRAN_SUFFIXES = {'.f', '.for', '.F', '.f90', '.F90'}
-LANG_SUFFIXES = {'.c', '.cc', '.cpp', '.cs', '.cu', '.d', '.m', '.mm', '.rs', '.java'} | FORTRAN_SUFFIXES
-LANG_SUPPORTED = {'c', 'cpp', 'cs', 'cuda', 'd', 'fortran', 'java', 'rust', 'objc', 'objcpp'}
+LANG_SUFFIXES = {'.c', '.cc', '.cpp', '.cs', '.cu', '.d', '.m', '.mm', '.rs', '.java', '.vala'} | FORTRAN_SUFFIXES
+LANG_SUPPORTED = {'c', 'cpp', 'cs', 'cuda', 'd', 'fortran', 'java', 'rust', 'objc', 'objcpp', 'vala'}
 
 DEFAULT_PROJECT = 'executable'
 DEFAULT_VERSION = '0.1'
@@ -53,12 +59,12 @@ meson compile -C builddir
 '''
 
 
-def create_sample(options: 'argparse.Namespace') -> None:
+def create_sample(options: Arguments) -> None:
     '''
     Based on what arguments are passed we check for a match in language
     then check for project type and create new Meson samples project.
     '''
-    sample_gen = sameple_generator(options)
+    sample_gen = sample_generator(options)
     if options.type == DEFAULT_TYPES['EXE'].value:
         sample_gen.create_executable()
     elif options.type == DEFAULT_TYPES['LIB'].value:
@@ -67,7 +73,7 @@ def create_sample(options: 'argparse.Namespace') -> None:
         raise RuntimeError('Unreachable code')
     print(INFO_MESSAGE)
 
-def autodetect_options(options: 'argparse.Namespace', sample: bool = False) -> None:
+def autodetect_options(options: Arguments, sample: bool = False) -> None:
     '''
     Here we autodetect options for args not passed in so don't have to
     think about it.
@@ -75,19 +81,20 @@ def autodetect_options(options: 'argparse.Namespace', sample: bool = False) -> N
     if not options.name:
         options.name = Path().resolve().stem
         if not re.match('[a-zA-Z_][a-zA-Z0-9]*', options.name) and sample:
-            raise SystemExit('Name of current directory "{}" is not usable as a sample project name.\n'
-                             'Specify a project name with --name.'.format(options.name))
-        print('Using "{}" (name of current directory) as project name.'
-              .format(options.name))
+            raise SystemExit(f'Name of current directory "{options.name}" is not usable as a sample project name.\n'
+                             'Specify a project name with --name.')
+        print(f'Using "{options.name}" (name of current directory) as project name.')
     if not options.executable:
         options.executable = options.name
-        print('Using "{}" (project name) as name of executable to build.'
-              .format(options.executable))
+        print(f'Using "{options.executable}" (project name) as name of executable to build.')
+    if options.executable in FORBIDDEN_TARGET_NAMES:
+        raise mesonlib.MesonException(f'Executable name {options.executable!r} is reserved for Meson internal use. '
+                                      'Refusing to init an invalid project.')
     if sample:
         # The rest of the autodetection is not applicable to generating sample projects.
         return
     if not options.srcfiles:
-        srcfiles = []
+        srcfiles: T.List[Path] = []
         for f in (f for f in Path().iterdir() if f.is_file()):
             if f.suffix in LANG_SUFFIXES:
                 srcfiles.append(f)
@@ -95,17 +102,16 @@ def autodetect_options(options: 'argparse.Namespace', sample: bool = False) -> N
             raise SystemExit('No recognizable source files found.\n'
                              'Run meson init in an empty directory to create a sample project.')
         options.srcfiles = srcfiles
-        print("Detected source files: " + ' '.join(map(str, srcfiles)))
-    options.srcfiles = [Path(f) for f in options.srcfiles]
+        print("Detected source files: " + ' '.join(str(s) for s in srcfiles))
     if not options.language:
         for f in options.srcfiles:
             if f.suffix == '.c':
                 options.language = 'c'
                 break
-            if f.suffix in ('.cc', '.cpp'):
+            if f.suffix in {'.cc', '.cpp'}:
                 options.language = 'cpp'
                 break
-            if f.suffix in '.cs':
+            if f.suffix == '.cs':
                 options.language = 'cs'
                 break
             if f.suffix == '.cu':
@@ -129,17 +135,23 @@ def autodetect_options(options: 'argparse.Namespace', sample: bool = False) -> N
             if f.suffix == '.java':
                 options.language = 'java'
                 break
+            if f.suffix == '.vala':
+                options.language = 'vala'
+                break
         if not options.language:
             raise SystemExit("Can't autodetect language, please specify it with -l.")
         print("Detected language: " + options.language)
 
+# Note: when adding arguments, please also add them to the completion
+# scripts in $MESONSRC/data/shell-completions/
 def add_arguments(parser: 'argparse.ArgumentParser') -> None:
     '''
     Here we add args for that the user can passed when making a new
     Meson project.
     '''
-    parser.add_argument("srcfiles", metavar="sourcefile", nargs="*", help="source files. default: all recognized files in current directory")
-    parser.add_argument('-C', default='.', dest='wd', help='directory to cd into before running')
+    parser.add_argument("srcfiles", metavar="sourcefile", nargs="*", type=Path, help="source files. default: all recognized files in current directory")
+    parser.add_argument('-C', dest='wd', action=mesonlib.RealPathAction,
+                        help='directory to cd into before running')
     parser.add_argument("-n", "--name", help="project name. default: name of current directory")
     parser.add_argument("-e", "--executable", help="executable name. default: project name")
     parser.add_argument("-d", "--deps", help="dependencies, comma-separated")
@@ -150,7 +162,7 @@ def add_arguments(parser: 'argparse.ArgumentParser') -> None:
     parser.add_argument('--type', default=DEFAULT_PROJECT, choices=('executable', 'library'), help=f"project type. default: {DEFAULT_PROJECT} based project")
     parser.add_argument('--version', default=DEFAULT_VERSION, help=f"project version. default: {DEFAULT_VERSION}")
 
-def run(options: 'argparse.Namespace') -> int:
+def run(options: Arguments) -> int:
     '''
     Here we generate the new Meson sample project.
     '''
@@ -174,10 +186,17 @@ def run(options: 'argparse.Namespace') -> int:
             print('Build directory already exists, deleting it.')
             shutil.rmtree(options.builddir)
         print('Building...')
-        cmd = mesonlib.get_meson_command() + [options.builddir]
+        cmd = mesonlib.get_meson_command() + ['setup', options.builddir]
         ret = subprocess.run(cmd)
         if ret.returncode:
             raise SystemExit
+
+        b = build.load(options.builddir)
+        need_vsenv = T.cast('bool', b.environment.coredata.get_option(mesonlib.OptionKey('vsenv')))
+        vsenv_active = mesonlib.setup_vsenv(need_vsenv)
+        if vsenv_active:
+            mlog.log(mlog.green('INFO:'), 'automatically activated MSVC compiler environment')
+
         cmd = detect_ninja() + ['-C', options.builddir]
         ret = subprocess.run(cmd)
         if ret.returncode:
